@@ -14,7 +14,11 @@
 
 #[cfg(feature = "sled_cryptostore")]
 use std::path::Path;
-use std::{collections::BTreeMap, mem, sync::Arc};
+use std::{
+    collections::{BTreeMap, HashSet},
+    mem,
+    sync::Arc,
+};
 
 use dashmap::DashMap;
 use matrix_sdk_common::{
@@ -34,8 +38,11 @@ use ruma::{
     },
     assign,
     events::{
-        room::encrypted::{EncryptedEventContent, EncryptedEventScheme},
+        room::encrypted::{
+            EncryptedEventContent, EncryptedEventScheme, EncryptedToDeviceEventContent,
+        },
         room_key::RoomKeyToDeviceEventContent,
+        secret::request::SecretName,
         AnyMessageEventContent, AnyRoomEvent, AnyToDeviceEvent, SyncMessageEvent, ToDeviceEvent,
     },
     DeviceId, DeviceIdBox, DeviceKeyAlgorithm, EventEncryptionAlgorithm, RoomId, UInt, UserId,
@@ -46,21 +53,21 @@ use tracing::{debug, error, info, trace, warn};
 use crate::store::sled::SledStore;
 use crate::{
     error::{EventError, MegolmError, MegolmResult, OlmError, OlmResult},
-    identities::{Device, IdentityManager, UserDevices},
-    key_request::KeyRequestMachine,
+    gossiping::GossipMachine,
+    identities::{user::UserIdentities, Device, IdentityManager, UserDevices},
     olm::{
-        Account, EncryptionSettings, ExportedRoomKey, GroupSessionKey, IdentityKeys,
-        InboundGroupSession, OlmDecryptionInfo, PrivateCrossSigningIdentity, ReadOnlyAccount,
-        SessionType,
+        Account, CrossSigningStatus, EncryptionSettings, ExportedRoomKey, GroupSessionKey,
+        IdentityKeys, InboundGroupSession, OlmDecryptionInfo, PrivateCrossSigningIdentity,
+        ReadOnlyAccount, SessionType,
     },
     requests::{IncomingResponse, OutgoingRequest, UploadSigningKeysRequest},
     session_manager::{GroupSessionManager, SessionManager},
     store::{
         Changes, CryptoStore, DeviceChanges, IdentityChanges, MemoryStore, Result as StoreResult,
-        Store,
+        SecretImportError, Store,
     },
     verification::{Verification, VerificationMachine, VerificationRequest},
-    ToDeviceRequest,
+    CrossSigningKeyExport, ToDeviceRequest,
 };
 
 /// State machine implementation of the Olm/Megolm encryption protocol used for
@@ -91,11 +98,10 @@ pub struct OlmMachine {
     verification_machine: VerificationMachine,
     /// The state machine that is responsible to handle outgoing and incoming
     /// key requests.
-    key_request_machine: KeyRequestMachine,
+    key_request_machine: GossipMachine,
     /// State machine handling public user identities and devices, keeping track
     /// of when a key query needs to be done and handling one.
     identity_manager: IdentityManager,
-    cross_signing_request: Arc<Mutex<Option<UploadSignaturesRequest>>>,
 }
 
 #[cfg(not(tarpaulin_include))]
@@ -155,7 +161,7 @@ impl OlmMachine {
 
         let group_session_manager = GroupSessionManager::new(account.clone(), store.clone());
 
-        let key_request_machine = KeyRequestMachine::new(
+        let key_request_machine = GossipMachine::new(
             user_id.clone(),
             device_id.clone(),
             store.clone(),
@@ -183,7 +189,6 @@ impl OlmMachine {
             verification_machine,
             key_request_machine,
             identity_manager,
-            cross_signing_request: Arc::new(Mutex::new(None)),
         }
     }
 
@@ -213,12 +218,15 @@ impl OlmMachine {
     ) -> StoreResult<Self> {
         let account = match store.load_account().await? {
             Some(a) => {
-                debug!("Restored account");
+                debug!(ed25519_key = a.identity_keys().ed25519(), "Restored an Olm account");
                 a
             }
             None => {
-                debug!("Creating a new account");
                 let account = ReadOnlyAccount::new(&user_id, &device_id);
+                debug!(
+                    ed25519_key = account.identity_keys().ed25519(),
+                    "Created a new Olm account"
+                );
                 store.save_account(account.clone()).await?;
                 account
             }
@@ -226,7 +234,14 @@ impl OlmMachine {
 
         let identity = match store.load_identity().await? {
             Some(i) => {
-                debug!("Restored the cross signing identity");
+                let master_key = i
+                    .master_public_key()
+                    .await
+                    .and_then(|m| m.get_first_key().map(|m| m.to_string()));
+                debug!(
+                    master_key =? master_key,
+                    "Restored the cross signing identity"
+                );
                 i
             }
             None => {
@@ -275,6 +290,16 @@ impl OlmMachine {
         self.account.identity_keys()
     }
 
+    /// Get the display name of our own device
+    pub async fn dislpay_name(&self) -> StoreResult<Option<String>> {
+        self.store.device_display_name().await
+    }
+
+    /// Get all the tracked users we know about
+    pub fn tracked_users(&self) -> HashSet<UserId> {
+        self.store.tracked_users()
+    }
+
     /// Get the outgoing requests that need to be sent out.
     ///
     /// This returns a list of `OutGoingRequest`, those requests need to be sent
@@ -534,7 +559,7 @@ impl OlmMachine {
     /// * `event` - The to-device event that should be decrypted.
     async fn decrypt_to_device_event(
         &self,
-        event: &ToDeviceEvent<EncryptedEventContent>,
+        event: &ToDeviceEvent<EncryptedToDeviceEventContent>,
     ) -> OlmResult<OlmDecryptionInfo> {
         let mut decrypted = self.account.decrypt_to_device_event(event).await?;
         // Handle the decrypted event, e.g. fetch out Megolm sessions out of
@@ -694,8 +719,12 @@ impl OlmMachine {
                 .key_request_machine
                 .receive_forwarded_room_key(&decrypted.sender_key, &mut e)
                 .await?),
+            AnyToDeviceEvent::SecretSend(mut e) => Ok((
+                self.key_request_machine.receive_secret(&decrypted.sender_key, &mut e).await?,
+                None,
+            )),
             _ => {
-                warn!("Received an unexpected encrypted to-device event");
+                warn!(event_type =? event.event_type(), "Received an unexpected encrypted to-device event");
                 Ok((Some(event), None))
             }
         }
@@ -736,8 +765,8 @@ impl OlmMachine {
         self.verification_machine.get_requests(user_id)
     }
 
-    async fn update_one_time_key_count(&self, key_count: &BTreeMap<DeviceKeyAlgorithm, UInt>) {
-        self.account.update_uploaded_key_count(key_count).await;
+    fn update_one_time_key_count(&self, key_count: &BTreeMap<DeviceKeyAlgorithm, UInt>) {
+        self.account.update_uploaded_key_count(key_count);
     }
 
     async fn handle_to_device_event(&self, event: &AnyToDeviceEvent) {
@@ -745,6 +774,9 @@ impl OlmMachine {
             AnyToDeviceEvent::RoomKeyRequest(e) => {
                 self.key_request_machine.receive_incoming_key_request(e)
             }
+            AnyToDeviceEvent::SecretRequest(e) => {
+                self.key_request_machine.receive_incoming_secret_request(e)
+            }
             AnyToDeviceEvent::KeyVerificationAccept(..)
             | AnyToDeviceEvent::KeyVerificationCancel(..)
             | AnyToDeviceEvent::KeyVerificationKey(..)
@@ -789,14 +821,14 @@ impl OlmMachine {
         one_time_keys_counts: &BTreeMap<DeviceKeyAlgorithm, UInt>,
     ) -> OlmResult<ToDevice> {
         // Remove verification objects that have expired or are done.
-        self.verification_machine.garbage_collect();
+        let mut events = self.verification_machine.garbage_collect();
 
         // Always save the account, a new session might get created which also
         // touches the account.
         let mut changes =
             Changes { account: Some(self.account.inner.clone()), ..Default::default() };
 
-        self.update_one_time_key_count(one_time_keys_counts).await;
+        self.update_one_time_key_count(one_time_keys_counts);
 
         for user_id in &changed_devices.changed {
             if let Err(e) = self.identity_manager.mark_user_as_changed(user_id).await {
@@ -804,26 +836,31 @@ impl OlmMachine {
             }
         }
 
-        let mut events = Vec::new();
-
         for mut raw_event in to_device_events.events {
             let event = match raw_event.deserialize() {
                 Ok(e) => e,
                 Err(e) => {
                     // Skip invalid events.
-                    warn!("Received an invalid to-device event {:?} {:?}", e, raw_event);
+                    warn!(
+                        error =? e,
+                        "Received an invalid to-device event"
+                    );
                     continue;
                 }
             };
 
-            info!("Received a to-device event {:?}", event);
+            info!(
+                sender = event.sender().as_str(),
+                event_type = event.event_type(),
+                "Received a to-device event"
+            );
 
             match event {
                 AnyToDeviceEvent::RoomEncrypted(e) => {
                     let decrypted = match self.decrypt_to_device_event(&e).await {
                         Ok(e) => e,
                         Err(err) => {
-                            warn!("Failed to decrypt to-device event from {} {}", e.sender, err);
+                            warn!(sender = e.sender.as_str(), error =? e, "Failed to decrypt to-device event");
 
                             if let OlmError::SessionWedged(sender, curve_key) = err {
                                 if let Err(e) = self
@@ -832,8 +869,9 @@ impl OlmMachine {
                                     .await
                                 {
                                     error!(
-                                        "Couldn't mark device from {} to be unwedged {:?}",
-                                        sender, e
+                                        sender = sender.as_str(),
+                                        error =? e,
+                                        "Couldn't mark device from to be unwedged",
                                     );
                                 }
                             }
@@ -928,7 +966,7 @@ impl OlmMachine {
                     .unwrap_or(false)
             }) {
             if (self.user_id() == device.user_id() && self.device_id() == device.device_id())
-                || device.is_trusted()
+                || device.verified()
             {
                 VerificationState::Trusted
             } else {
@@ -1054,6 +1092,18 @@ impl OlmMachine {
         self.store.get_device(user_id, device_id).await
     }
 
+    /// Get the cross signing user identity of a user.
+    ///
+    /// # Arguments
+    ///
+    /// * `user_id` - The unique id of the user that the identity belongs to
+    ///
+    /// Returns a `UserIdentities` enum if one is found and the crypto store
+    /// didn't throw an error.
+    pub async fn get_identity(&self, user_id: &UserId) -> StoreResult<Option<UserIdentities>> {
+        self.store.get_identity(user_id).await
+    }
+
     /// Get a map holding all the devices of an user.
     ///
     /// # Arguments
@@ -1215,6 +1265,46 @@ impl OlmMachine {
 
         Ok(exported)
     }
+
+    /// Get the status of the private cross signing keys.
+    ///
+    /// This can be used to check which private cross signing keys we have
+    /// stored locally.
+    pub async fn cross_signing_status(&self) -> CrossSigningStatus {
+        self.user_identity.lock().await.status().await
+    }
+
+    /// Export all the private cross signing keys we have.
+    ///
+    /// The export will contain the seed for the ed25519 keys as a unpadded
+    /// base64 encoded string.
+    ///
+    /// This method returns `None` if we don't have any private cross signing
+    /// keys.
+    pub async fn export_cross_signing_keys(&self) -> Option<CrossSigningKeyExport> {
+        let master_key = self.store.export_secret(&SecretName::CrossSigningMasterKey).await;
+        let self_signing_key =
+            self.store.export_secret(&SecretName::CrossSigningSelfSigningKey).await;
+        let user_signing_key =
+            self.store.export_secret(&SecretName::CrossSigningUserSigningKey).await;
+
+        if master_key.is_none() && self_signing_key.is_none() && user_signing_key.is_none() {
+            None
+        } else {
+            Some(CrossSigningKeyExport { master_key, self_signing_key, user_signing_key })
+        }
+    }
+
+    /// Import our private cross signing keys.
+    ///
+    /// The export needs to contain the seed for the ed25519 keys as an unpadded
+    /// base64 encoded string.
+    pub async fn import_cross_signing_keys(
+        &self,
+        export: CrossSigningKeyExport,
+    ) -> Result<CrossSigningStatus, SecretImportError> {
+        self.store.import_cross_signing_keys(export).await
+    }
 }
 
 #[cfg(test)]
@@ -1235,19 +1325,18 @@ pub(crate) mod test {
             IncomingResponse,
         },
         encryption::OneTimeKey,
+        event_id,
         events::{
             dummy::DummyToDeviceEventContent,
             room::{
-                encrypted::EncryptedEventContent,
+                encrypted::EncryptedToDeviceEventContent,
                 message::{MessageEventContent, MessageType},
             },
             AnyMessageEventContent, AnySyncMessageEvent, AnySyncRoomEvent, AnyToDeviceEvent,
             AnyToDeviceEventContent, SyncMessageEvent, ToDeviceEvent, Unsigned,
         },
-        identifiers::{
-            event_id, room_id, user_id, DeviceId, DeviceKeyAlgorithm, DeviceKeyId, UserId,
-        },
-        uint, MilliSecondsSinceUnixEpoch,
+        room_id, uint, user_id, DeviceId, DeviceKeyAlgorithm, DeviceKeyId,
+        MilliSecondsSinceUnixEpoch, UserId,
     };
     use serde_json::json;
 
@@ -1289,7 +1378,9 @@ pub(crate) mod test {
             .expect("Can't parse the keys upload response")
     }
 
-    fn to_device_requests_to_content(requests: Vec<Arc<ToDeviceRequest>>) -> EncryptedEventContent {
+    fn to_device_requests_to_content(
+        requests: Vec<Arc<ToDeviceRequest>>,
+    ) -> EncryptedToDeviceEventContent {
         let to_device_request = &requests[0];
 
         to_device_request
@@ -1401,6 +1492,10 @@ pub(crate) mod test {
         response.one_time_key_counts.insert(DeviceKeyAlgorithm::SignedCurve25519, uint!(50));
         machine.receive_keys_upload_response(&response).await.unwrap();
         assert!(!machine.should_upload_keys().await);
+
+        response.one_time_key_counts.remove(&DeviceKeyAlgorithm::SignedCurve25519);
+        machine.receive_keys_upload_response(&response).await.unwrap();
+        assert!(!machine.should_upload_keys().await);
     }
 
     #[tokio::test]
@@ -1772,7 +1867,7 @@ pub(crate) mod test {
 
         let bob_device = alice.get_device(bob.user_id(), bob.device_id()).await.unwrap().unwrap();
 
-        assert!(!bob_device.is_trusted());
+        assert!(!bob_device.verified());
 
         let (alice_sas, request) = bob_device.start_verification().await.unwrap();
 
@@ -1835,14 +1930,14 @@ pub(crate) mod test {
             .unwrap();
 
         assert!(alice_sas.is_done());
-        assert!(bob_device.is_trusted());
+        assert!(bob_device.verified());
 
         let alice_device =
             bob.get_device(alice.user_id(), alice.device_id()).await.unwrap().unwrap();
 
-        assert!(!alice_device.is_trusted());
+        assert!(!alice_device.verified());
         bob.handle_verification_event(&event).await;
         assert!(bob_sas.is_done());
-        assert!(alice_device.is_trusted());
+        assert!(alice_device.verified());
     }
 }
