@@ -15,14 +15,21 @@
 mod cache;
 mod event_enums;
 mod machine;
+#[cfg(feature = "qrcode")]
 mod qrcode;
 mod requests;
 mod sas;
 
-use std::sync::Arc;
+use std::{
+    collections::{BTreeMap, HashMap},
+    sync::Arc,
+};
 
 use event_enums::OutgoingContent;
 pub use machine::VerificationMachine;
+use matrix_sdk_common::locks::Mutex;
+#[cfg(feature = "qrcode")]
+#[cfg_attr(feature = "docs", doc(cfg(qrcode)))]
 pub use qrcode::QrVerification;
 pub use requests::VerificationRequest;
 use ruma::{
@@ -35,23 +42,84 @@ use ruma::{
         },
         AnyMessageEventContent, AnyToDeviceEventContent,
     },
-    DeviceId, EventId, RoomId, UserId,
+    DeviceId, DeviceIdBox, DeviceKeyId, EventId, RoomId, UserId,
 };
 pub use sas::{AcceptSettings, Sas};
 use tracing::{error, info, trace, warn};
 
 use crate::{
     error::SignatureError,
-    olm::PrivateCrossSigningIdentity,
+    gossiping::{GossipMachine, GossipRequest},
+    olm::{PrivateCrossSigningIdentity, ReadOnlyAccount, Session},
     store::{Changes, CryptoStore},
-    CryptoStoreError, LocalTrust, ReadOnlyDevice, UserIdentities,
+    CryptoStoreError, LocalTrust, ReadOnlyDevice, ReadOnlyUserIdentities,
 };
 
+#[derive(Clone, Debug)]
+pub(crate) struct VerificationStore {
+    pub account: ReadOnlyAccount,
+    inner: Arc<dyn CryptoStore>,
+}
+
+impl VerificationStore {
+    pub async fn get_device(
+        &self,
+        user_id: &UserId,
+        device_id: &DeviceId,
+    ) -> Result<Option<ReadOnlyDevice>, CryptoStoreError> {
+        Ok(self.inner.get_device(user_id, device_id).await?.filter(|d| {
+            !(d.user_id() == self.account.user_id() && d.device_id() == self.account.device_id())
+        }))
+    }
+
+    pub async fn get_user_identity(
+        &self,
+        user_id: &UserId,
+    ) -> Result<Option<ReadOnlyUserIdentities>, CryptoStoreError> {
+        self.inner.get_user_identity(user_id).await
+    }
+
+    pub async fn save_changes(&self, changes: Changes) -> Result<(), CryptoStoreError> {
+        self.inner.save_changes(changes).await
+    }
+
+    pub async fn get_user_devices(
+        &self,
+        user_id: &UserId,
+    ) -> Result<HashMap<DeviceIdBox, ReadOnlyDevice>, CryptoStoreError> {
+        self.inner.get_user_devices(user_id).await
+    }
+
+    pub async fn get_sessions(
+        &self,
+        sender_key: &str,
+    ) -> Result<Option<Arc<Mutex<Vec<Session>>>>, CryptoStoreError> {
+        self.inner.get_sessions(sender_key).await
+    }
+
+    /// Get the signatures that have signed our own device.
+    pub async fn device_signatures(
+        &self,
+    ) -> Result<Option<BTreeMap<UserId, BTreeMap<DeviceKeyId, String>>>, CryptoStoreError> {
+        Ok(self
+            .inner
+            .get_device(self.account.user_id(), self.account.device_id())
+            .await?
+            .map(|d| d.signatures().to_owned()))
+    }
+
+    pub fn inner(&self) -> &dyn CryptoStore {
+        &*self.inner
+    }
+}
+
 /// An enum over the different verification types the SDK supports.
 #[derive(Clone, Debug)]
 pub enum Verification {
     /// The `m.sas.v1` verification variant.
     SasV1(Sas),
+    #[cfg(feature = "qrcode")]
+    #[cfg_attr(feature = "docs", doc(cfg(qrcode)))]
     /// The `m.qr_code.*.v1` verification variant.
     QrV1(QrVerification),
 }
@@ -59,6 +127,7 @@ pub enum Verification {
 impl Verification {
     /// Try to deconstruct this verification enum into a SAS verification.
     pub fn sas_v1(self) -> Option<Sas> {
+        #[allow(irrefutable_let_patterns)]
         if let Verification::SasV1(sas) = self {
             Some(sas)
         } else {
@@ -66,6 +135,8 @@ impl Verification {
         }
     }
 
+    #[cfg(feature = "qrcode")]
+    #[cfg_attr(feature = "docs", doc(cfg(qrcode)))]
     /// Try to deconstruct this verification enum into a QR code verification.
     pub fn qr_v1(self) -> Option<QrVerification> {
         if let Verification::QrV1(qr) = self {
@@ -79,6 +150,7 @@ impl Verification {
     pub fn is_done(&self) -> bool {
         match self {
             Verification::SasV1(s) => s.is_done(),
+            #[cfg(feature = "qrcode")]
             Verification::QrV1(qr) => qr.is_done(),
         }
     }
@@ -87,6 +159,7 @@ impl Verification {
     pub fn flow_id(&self) -> &str {
         match self {
             Verification::SasV1(s) => s.flow_id().as_str(),
+            #[cfg(feature = "qrcode")]
             Verification::QrV1(qr) => qr.flow_id().as_str(),
         }
     }
@@ -95,6 +168,7 @@ impl Verification {
     pub fn is_cancelled(&self) -> bool {
         match self {
             Verification::SasV1(s) => s.is_cancelled(),
+            #[cfg(feature = "qrcode")]
             Verification::QrV1(qr) => qr.is_cancelled(),
         }
     }
@@ -103,6 +177,7 @@ impl Verification {
     pub fn user_id(&self) -> &UserId {
         match self {
             Verification::SasV1(v) => v.user_id(),
+            #[cfg(feature = "qrcode")]
             Verification::QrV1(v) => v.user_id(),
         }
     }
@@ -111,6 +186,7 @@ impl Verification {
     pub fn other_user(&self) -> &UserId {
         match self {
             Verification::SasV1(s) => s.other_user_id(),
+            #[cfg(feature = "qrcode")]
             Verification::QrV1(qr) => qr.other_user_id(),
         }
     }
@@ -119,6 +195,7 @@ impl Verification {
     pub fn is_self_verification(&self) -> bool {
         match self {
             Verification::SasV1(v) => v.is_self_verification(),
+            #[cfg(feature = "qrcode")]
             Verification::QrV1(v) => v.is_self_verification(),
         }
     }
@@ -130,6 +207,8 @@ impl From<Sas> for Verification {
     }
 }
 
+#[cfg(feature = "qrcode")]
+#[cfg_attr(feature = "docs", doc(cfg(qrcode)))]
 impl From<QrVerification> for Verification {
     fn from(qr: QrVerification) -> Self {
         Self::QrV1(qr)
@@ -144,7 +223,7 @@ impl From<QrVerification> for Verification {
 #[derive(Clone, Debug)]
 pub struct Done {
     verified_devices: Arc<[ReadOnlyDevice]>,
-    verified_master_keys: Arc<[UserIdentities]>,
+    verified_master_keys: Arc<[ReadOnlyUserIdentities]>,
 }
 
 impl Done {
@@ -165,14 +244,47 @@ impl Done {
     }
 }
 
+/// Information about the cancellation of a verification request or verification
+/// flow.
+#[derive(Clone, Debug)]
+pub struct CancelInfo {
+    cancelled_by_us: bool,
+    cancel_code: CancelCode,
+    reason: &'static str,
+}
+
+impl CancelInfo {
+    /// Get the human readable reason of the cancellation.
+    pub fn reason(&self) -> &'static str {
+        self.reason
+    }
+
+    /// Get the `CancelCode` that cancelled this verification.
+    pub fn cancel_code(&self) -> &CancelCode {
+        &self.cancel_code
+    }
+
+    /// Was the verification cancelled by us?
+    pub fn cancelled_by_us(&self) -> bool {
+        self.cancelled_by_us
+    }
+}
+
+impl From<Cancelled> for CancelInfo {
+    fn from(c: Cancelled) -> Self {
+        Self { cancelled_by_us: c.cancelled_by_us, cancel_code: c.cancel_code, reason: c.reason }
+    }
+}
+
 #[derive(Clone, Debug)]
 pub struct Cancelled {
+    cancelled_by_us: bool,
     cancel_code: CancelCode,
     reason: &'static str,
 }
 
 impl Cancelled {
-    fn new(code: CancelCode) -> Self {
+    fn new(cancelled_by_us: bool, code: CancelCode) -> Self {
         let reason = match code {
             CancelCode::Accepted => {
                 "A m.key.verification.request was accepted by a different device."
@@ -192,7 +304,7 @@ impl Cancelled {
             _ => "Unknown cancel reason",
         };
 
-        Self { cancel_code: code, reason }
+        Self { cancelled_by_us, cancel_code: code, reason }
     }
 
     pub fn as_content(&self, flow_id: &FlowId) -> OutgoingContent {
@@ -274,12 +386,13 @@ pub enum VerificationResult {
 #[derive(Clone, Debug)]
 pub struct IdentitiesBeingVerified {
     private_identity: PrivateCrossSigningIdentity,
-    store: Arc<dyn CryptoStore>,
+    store: VerificationStore,
     device_being_verified: ReadOnlyDevice,
-    identity_being_verified: Option<UserIdentities>,
+    identity_being_verified: Option<ReadOnlyUserIdentities>,
 }
 
 impl IdentitiesBeingVerified {
+    #[cfg(feature = "qrcode")]
     async fn can_sign_devices(&self) -> bool {
         self.private_identity.can_sign_devices().await
     }
@@ -307,10 +420,11 @@ impl IdentitiesBeingVerified {
     pub async fn mark_as_done(
         &self,
         verified_devices: Option<&[ReadOnlyDevice]>,
-        verified_identities: Option<&[UserIdentities]>,
+        verified_identities: Option<&[ReadOnlyUserIdentities]>,
     ) -> Result<VerificationResult, CryptoStoreError> {
         let device = self.mark_device_as_verified(verified_devices).await?;
-        let identity = self.mark_identity_as_verified(verified_identities).await?;
+        let (identity, should_request_secrets) =
+            self.mark_identity_as_verified(verified_identities).await?;
 
         if device.is_none() && identity.is_none() {
             // Something wen't wrong if nothing was verified, we use key
@@ -328,7 +442,7 @@ impl IdentitiesBeingVerified {
                     Err(SignatureError::MissingSigningKey) => {
                         warn!(
                             "Can't sign the device keys for {} {}, \
-                                  no private user signing key found",
+                                  no private device signing key found",
                             device.user_id(),
                             device.device_id(),
                         );
@@ -404,6 +518,11 @@ impl IdentitiesBeingVerified {
             identity_signature_request
         };
 
+        if should_request_secrets {
+            let secret_requests = self.request_missing_secrets().await;
+            changes.key_requests = secret_requests;
+        }
+
         // TODO store the signature upload request as well.
         self.store.save_changes(changes).await?;
 
@@ -412,19 +531,24 @@ impl IdentitiesBeingVerified {
             .unwrap_or(VerificationResult::Ok))
     }
 
+    async fn request_missing_secrets(&self) -> Vec<GossipRequest> {
+        let secrets = self.private_identity.get_missing_secrets().await;
+        GossipMachine::request_missing_secrets(self.user_id(), secrets)
+    }
+
     async fn mark_identity_as_verified(
         &self,
-        verified_identities: Option<&[UserIdentities]>,
-    ) -> Result<Option<UserIdentities>, CryptoStoreError> {
+        verified_identities: Option<&[ReadOnlyUserIdentities]>,
+    ) -> Result<(Option<ReadOnlyUserIdentities>, bool), CryptoStoreError> {
         // If there wasn't an identity available during the verification flow
         // return early as there's nothing to do.
         if self.identity_being_verified.is_none() {
-            return Ok(None);
+            return Ok((None, false));
         }
 
         let identity = self.store.get_user_identity(self.other_user_id()).await?;
 
-        if let Some(identity) = identity {
+        Ok(if let Some(identity) = identity {
             if self
                 .identity_being_verified
                 .as_ref()
@@ -436,11 +560,14 @@ impl IdentitiesBeingVerified {
                         "Marking the user identity of as verified."
                     );
 
-                    if let UserIdentities::Own(i) = &identity {
+                    let should_request_secrets = if let ReadOnlyUserIdentities::Own(i) = &identity {
                         i.mark_as_verified();
-                    }
+                        true
+                    } else {
+                        false
+                    };
 
-                    Ok(Some(identity))
+                    (Some(identity), should_request_secrets)
                 } else {
                     info!(
                         user_id = self.other_user_id().as_str(),
@@ -449,7 +576,7 @@ impl IdentitiesBeingVerified {
                          the interactive verification",
                     );
 
-                    Ok(None)
+                    (None, false)
                 }
             } else {
                 warn!(
@@ -458,7 +585,7 @@ impl IdentitiesBeingVerified {
                       verification was going on, not marking the identity as verified.",
                 );
 
-                Ok(None)
+                (None, false)
             }
         } else {
             info!(
@@ -466,8 +593,8 @@ impl IdentitiesBeingVerified {
                 "The identity of the user was deleted while an interactive \
                  verification was going on.",
             );
-            Ok(None)
-        }
+            (None, false)
+        })
     }
 
     async fn mark_device_as_verified(
@@ -524,6 +651,8 @@ impl IdentitiesBeingVerified {
 
 #[cfg(test)]
 pub(crate) mod test {
+    use std::convert::TryInto;
+
     use ruma::{
         events::{AnyToDeviceEvent, AnyToDeviceEventContent, ToDeviceEvent},
         UserId,
@@ -539,7 +668,8 @@ pub(crate) mod test {
         sender: &UserId,
         request: &OutgoingVerificationRequest,
     ) -> AnyToDeviceEvent {
-        let content = request.to_owned().into();
+        let content =
+            request.to_owned().try_into().expect("Can't fetch content out of the request");
         wrap_any_to_device_content(sender, content)
     }
 
