@@ -33,19 +33,20 @@ use ruma::{
         },
         AnyMessageEventContent, AnyToDeviceEventContent,
     },
-    identifiers::{DeviceIdBox, DeviceKeyAlgorithm, UserId},
+    DeviceId, DeviceIdBox, DeviceKeyAlgorithm, RoomId, UserId,
 };
 use thiserror::Error;
+use tracing::trace;
 
 use super::{
-    event_enums::{DoneContent, OutgoingContent, OwnedStartContent, StartContent},
-    Cancelled, Done, FlowId, IdentitiesBeingVerified, VerificationResult,
+    event_enums::{CancelContent, DoneContent, OutgoingContent, OwnedStartContent, StartContent},
+    requests::RequestHandle,
+    CancelInfo, Cancelled, Done, FlowId, IdentitiesBeingVerified, VerificationResult,
+    VerificationStore,
 };
 use crate::{
-    olm::{PrivateCrossSigningIdentity, ReadOnlyAccount},
-    store::CryptoStore,
-    CryptoStoreError, OutgoingVerificationRequest, ReadOnlyDevice, RoomMessageRequest,
-    ToDeviceRequest, UserIdentities,
+    olm::PrivateCrossSigningIdentity, CryptoStoreError, OutgoingVerificationRequest,
+    ReadOnlyDevice, ReadOnlyUserIdentities, RoomMessageRequest, ToDeviceRequest,
 };
 
 const SECRET_SIZE: usize = 16;
@@ -79,10 +80,12 @@ pub enum ScanError {
 #[derive(Clone)]
 pub struct QrVerification {
     flow_id: FlowId,
-    store: Arc<dyn CryptoStore>,
+    store: VerificationStore,
     inner: Arc<QrVerificationData>,
     state: Arc<Mutex<InnerState>>,
     identities: IdentitiesBeingVerified,
+    request_handle: Option<RequestHandle>,
+    we_started: bool,
 }
 
 impl std::fmt::Debug for QrVerification {
@@ -100,10 +103,15 @@ impl QrVerification {
     ///
     /// When the verification object is in this state it's required that the
     /// user confirms that the other side has scanned the QR code.
-    pub fn is_scanned(&self) -> bool {
+    pub fn has_been_scanned(&self) -> bool {
         matches!(&*self.state.lock().unwrap(), InnerState::Scanned(_))
     }
 
+    /// Has the scanning of the QR code been confirmed by us.
+    pub fn has_been_confirmed(&self) -> bool {
+        matches!(&*self.state.lock().unwrap(), InnerState::Confirmed(_))
+    }
+
     /// Get our own user id.
     pub fn user_id(&self) -> &UserId {
         self.identities.user_id()
@@ -115,6 +123,26 @@ impl QrVerification {
         self.identities.other_user_id()
     }
 
+    /// Get the device id of the other side.
+    pub fn other_device_id(&self) -> &DeviceId {
+        self.identities.other_device_id()
+    }
+
+    /// Did we initiate the verification request
+    pub fn we_started(&self) -> bool {
+        self.we_started
+    }
+
+    /// Get info about the cancellation if the verification flow has been
+    /// cancelled.
+    pub fn cancel_info(&self) -> Option<CancelInfo> {
+        if let InnerState::Cancelled(c) = &*self.state.lock().unwrap() {
+            Some(c.state.clone().into())
+        } else {
+            None
+        }
+    }
+
     /// Has the verification flow completed.
     pub fn is_done(&self) -> bool {
         matches!(&*self.state.lock().unwrap(), InnerState::Done(_))
@@ -130,11 +158,25 @@ impl QrVerification {
         self.identities.is_self_verification()
     }
 
+    /// Have we successfully scanned the QR code and are able to send a
+    /// reciprocation event.
+    pub fn reciprocated(&self) -> bool {
+        matches!(&*self.state.lock().unwrap(), InnerState::Reciprocated(_))
+    }
+
     /// Get the unique ID that identifies this QR code verification flow.
     pub fn flow_id(&self) -> &FlowId {
         &self.flow_id
     }
 
+    /// Get the room id if the verification is happening inside a room.
+    pub fn room_id(&self) -> Option<&RoomId> {
+        match self.flow_id() {
+            FlowId::ToDevice(_) => None,
+            FlowId::InRoom(r, _) => Some(r),
+        }
+    }
+
     /// Generate a QR code object that is representing this verification flow.
     ///
     /// The `QrCode` can then be rendered as an image or as an unicode string.
@@ -156,7 +198,43 @@ impl QrVerification {
 
     /// Cancel the verification flow.
     pub fn cancel(&self) -> Option<OutgoingVerificationRequest> {
-        self.cancel_with_code(CancelCode::User).map(|c| self.content_to_request(c))
+        self.cancel_with_code(CancelCode::User)
+    }
+
+    /// Cancel the verification.
+    ///
+    /// This cancels the verification with given `CancelCode`.
+    ///
+    /// **Note**: This method should generally not be used, the [`cancel()`]
+    /// method should be preferred. The SDK will automatically cancel with the
+    /// approprate cancel code, user initiated cancellations should only cancel
+    /// with the `CancelCode::User`
+    ///
+    /// Returns None if the `Sas` object is already in a canceled state,
+    /// otherwise it returns a request that needs to be sent out.
+    ///
+    /// [`cancel()`]: #method.cancel
+    pub fn cancel_with_code(&self, code: CancelCode) -> Option<OutgoingVerificationRequest> {
+        let mut state = self.state.lock().unwrap();
+
+        if let Some(request) = &self.request_handle {
+            request.cancel_with_code(&code);
+        }
+
+        let new_state = QrState::<Cancelled>::new(true, code);
+        let content = new_state.as_content(self.flow_id());
+
+        match &*state {
+            InnerState::Confirmed(_)
+            | InnerState::Created(_)
+            | InnerState::Scanned(_)
+            | InnerState::Reciprocated(_)
+            | InnerState::Done(_) => {
+                *state = InnerState::Cancelled(new_state);
+                Some(self.content_to_request(content))
+            }
+            InnerState::Cancelled(_) => None,
+        }
     }
 
     /// Notify the other side that we have successfully scanned the QR code and
@@ -164,9 +242,11 @@ impl QrVerification {
     ///
     /// This will return some `OutgoingContent` if the object is in the correct
     /// state to start the verification flow, otherwise `None`.
-    pub fn reciprocate(&self) -> Option<OutgoingContent> {
+    pub fn reciprocate(&self) -> Option<OutgoingVerificationRequest> {
         match &*self.state.lock().unwrap() {
-            InnerState::Reciprocated(s) => Some(s.as_content(self.flow_id())),
+            InnerState::Reciprocated(s) => {
+                Some(self.content_to_request(s.as_content(self.flow_id())))
+            }
             InnerState::Created(_)
             | InnerState::Scanned(_)
             | InnerState::Confirmed(_)
@@ -209,25 +289,6 @@ impl QrVerification {
         }
     }
 
-    fn cancel_with_code(&self, code: CancelCode) -> Option<OutgoingContent> {
-        let new_state = QrState::<Cancelled>::new(code);
-        let content = new_state.as_content(self.flow_id());
-
-        let mut state = self.state.lock().unwrap();
-
-        match &*state {
-            InnerState::Confirmed(_)
-            | InnerState::Created(_)
-            | InnerState::Scanned(_)
-            | InnerState::Reciprocated(_)
-            | InnerState::Done(_) => {
-                *state = InnerState::Cancelled(new_state);
-                Some(content)
-            }
-            InnerState::Cancelled(_) => None,
-        }
-    }
-
     async fn mark_as_done(
         &self,
         new_state: QrState<Done>,
@@ -243,7 +304,7 @@ impl QrVerification {
             match self.identities.mark_as_done(Some(&devices), Some(&identities)).await? {
                 VerificationResult::Ok => (None, None),
                 VerificationResult::Cancel(c) => {
-                    let canceled = QrState::<Cancelled>::new(c);
+                    let canceled = QrState::<Cancelled>::new(false, c);
                     let content = canceled.as_content(self.flow_id());
                     new_state = InnerState::Cancelled(canceled);
                     (Some(content), None)
@@ -338,6 +399,28 @@ impl QrVerification {
         }
     }
 
+    pub(crate) fn receive_cancel(&self, sender: &UserId, content: &CancelContent<'_>) {
+        if sender == self.other_user_id() {
+            let mut state = self.state.lock().unwrap();
+
+            let new_state = match &*state {
+                InnerState::Created(s) => s.clone().into_cancelled(content),
+                InnerState::Scanned(s) => s.clone().into_cancelled(content),
+                InnerState::Confirmed(s) => s.clone().into_cancelled(content),
+                InnerState::Reciprocated(s) => s.clone().into_cancelled(content),
+                InnerState::Done(_) | InnerState::Cancelled(_) => return,
+            };
+
+            trace!(
+                sender = sender.as_str(),
+                code = content.cancel_code().as_str(),
+                "Cancelling a QR verification, other user has cancelled"
+            );
+
+            *state = InnerState::Cancelled(new_state);
+        }
+    }
+
     fn generate_secret() -> String {
         let mut shared_secret = [0u8; SECRET_SIZE];
         getrandom::getrandom(&mut shared_secret)
@@ -346,11 +429,13 @@ impl QrVerification {
     }
 
     pub(crate) fn new_self(
-        store: Arc<dyn CryptoStore>,
+        store: VerificationStore,
         flow_id: FlowId,
         own_master_key: String,
         other_device_key: String,
         identities: IdentitiesBeingVerified,
+        we_started: bool,
+        request_handle: Option<RequestHandle>,
     ) -> Self {
         let secret = Self::generate_secret();
 
@@ -362,35 +447,38 @@ impl QrVerification {
         )
         .into();
 
-        Self::new_helper(store, flow_id, inner, identities)
+        Self::new_helper(store, flow_id, inner, identities, we_started, request_handle)
     }
 
     pub(crate) fn new_self_no_master(
-        account: ReadOnlyAccount,
-        store: Arc<dyn CryptoStore>,
+        store: VerificationStore,
         flow_id: FlowId,
         own_master_key: String,
         identities: IdentitiesBeingVerified,
+        we_started: bool,
+        request_handle: Option<RequestHandle>,
     ) -> QrVerification {
         let secret = Self::generate_secret();
 
         let inner: QrVerificationData = SelfVerificationNoMasterKey::new(
             flow_id.as_str().to_owned(),
-            account.identity_keys().ed25519().to_string(),
+            store.account.identity_keys().ed25519().to_string(),
             own_master_key,
             secret,
         )
         .into();
 
-        Self::new_helper(store, flow_id, inner, identities)
+        Self::new_helper(store, flow_id, inner, identities, we_started, request_handle)
     }
 
     pub(crate) fn new_cross(
-        store: Arc<dyn CryptoStore>,
+        store: VerificationStore,
         flow_id: FlowId,
         own_master_key: String,
         other_master_key: String,
         identities: IdentitiesBeingVerified,
+        we_started: bool,
+        request_handle: Option<RequestHandle>,
     ) -> Self {
         let secret = Self::generate_secret();
 
@@ -403,17 +491,19 @@ impl QrVerification {
         let inner: QrVerificationData =
             VerificationData::new(event_id, own_master_key, other_master_key, secret).into();
 
-        Self::new_helper(store, flow_id, inner, identities)
+        Self::new_helper(store, flow_id, inner, identities, we_started, request_handle)
     }
 
+    #[allow(clippy::too_many_arguments)]
     pub(crate) async fn from_scan(
-        store: Arc<dyn CryptoStore>,
-        own_account: ReadOnlyAccount,
+        store: VerificationStore,
         private_identity: PrivateCrossSigningIdentity,
         other_user_id: UserId,
         other_device_id: DeviceIdBox,
         flow_id: FlowId,
         qr_code: QrVerificationData,
+        we_started: bool,
+        request_handle: Option<RequestHandle>,
     ) -> Result<Self, ScanError> {
         if flow_id.as_str() != qr_code.flow_id() {
             return Err(ScanError::FlowIdMismatch {
@@ -423,8 +513,8 @@ impl QrVerification {
         }
 
         let own_identity =
-            store.get_user_identity(own_account.user_id()).await?.ok_or_else(|| {
-                ScanError::MissingCrossSigningIdentity(own_account.user_id().to_owned())
+            store.get_user_identity(store.account.user_id()).await?.ok_or_else(|| {
+                ScanError::MissingCrossSigningIdentity(store.account.user_id().to_owned())
             })?;
         let other_identity = store
             .get_user_identity(&other_user_id)
@@ -435,7 +525,7 @@ impl QrVerification {
                 ScanError::MissingDeviceKeys(other_user_id.clone(), other_device_id.clone())
             })?;
 
-        let check_master_key = |key, identity: &UserIdentities| {
+        let check_master_key = |key, identity: &ReadOnlyUserIdentities| {
             let master_key = identity.master_key().get_first_key().ok_or_else(|| {
                 ScanError::MissingCrossSigningIdentity(identity.user_id().clone())
             })?;
@@ -450,33 +540,23 @@ impl QrVerification {
             }
         };
 
-        let identities = match qr_code {
+        let (device_being_verified, identity_being_verified) = match qr_code {
             QrVerificationData::Verification(_) => {
                 check_master_key(qr_code.first_key(), &other_identity)?;
                 check_master_key(qr_code.second_key(), &own_identity)?;
 
-                IdentitiesBeingVerified {
-                    private_identity,
-                    store: store.clone(),
-                    device_being_verified: other_device,
-                    identity_being_verified: Some(other_identity),
-                }
+                (other_device, Some(other_identity))
             }
             QrVerificationData::SelfVerification(_) => {
                 check_master_key(qr_code.first_key(), &other_identity)?;
-                if qr_code.second_key() != own_account.identity_keys().ed25519() {
+                if qr_code.second_key() != store.account.identity_keys().ed25519() {
                     return Err(ScanError::KeyMismatch {
-                        expected: own_account.identity_keys().ed25519().to_owned(),
+                        expected: store.account.identity_keys().ed25519().to_owned(),
                         found: qr_code.second_key().to_owned(),
                     });
                 }
 
-                IdentitiesBeingVerified {
-                    private_identity,
-                    store: store.clone(),
-                    device_being_verified: other_device,
-                    identity_being_verified: Some(other_identity),
-                }
+                (other_device, Some(other_identity))
             }
             QrVerificationData::SelfVerificationNoMasterKey(_) => {
                 let device_key =
@@ -490,34 +570,41 @@ impl QrVerification {
                     });
                 }
                 check_master_key(qr_code.second_key(), &other_identity)?;
-                IdentitiesBeingVerified {
-                    private_identity,
-                    store: store.clone(),
-                    device_being_verified: other_device,
-                    identity_being_verified: None,
-                }
+                (other_device, Some(other_identity))
             }
         };
 
+        let identities = IdentitiesBeingVerified {
+            private_identity,
+            store: store.clone(),
+            device_being_verified,
+            identity_being_verified,
+        };
+
         let secret = qr_code.secret().to_owned();
+        let own_device_id = store.account.device_id().to_owned();
 
         Ok(Self {
             store,
             flow_id,
             inner: qr_code.into(),
             state: Mutex::new(InnerState::Reciprocated(QrState {
-                state: Reciprocated { secret, own_device_id: own_account.device_id().to_owned() },
+                state: Reciprocated { secret, own_device_id },
             }))
             .into(),
             identities,
+            we_started,
+            request_handle,
         })
     }
 
     fn new_helper(
-        store: Arc<dyn CryptoStore>,
+        store: VerificationStore,
         flow_id: FlowId,
         inner: QrVerificationData,
         identities: IdentitiesBeingVerified,
+        we_started: bool,
+        request_handle: Option<RequestHandle>,
     ) -> Self {
         let secret = inner.secret().to_owned();
 
@@ -527,6 +614,8 @@ impl QrVerification {
             inner: inner.into(),
             state: Mutex::new(InnerState::Created(QrState { state: Created { secret } })).into(),
             identities,
+            we_started,
+            request_handle,
         }
     }
 }
@@ -546,6 +635,12 @@ struct QrState<S: Clone> {
     state: S,
 }
 
+impl<S: Clone> QrState<S> {
+    pub fn into_cancelled(self, content: &CancelContent<'_>) -> QrState<Cancelled> {
+        QrState { state: Cancelled::new(false, content.cancel_code().to_owned()) }
+    }
+}
+
 #[derive(Clone, Debug)]
 struct Created {
     secret: String,
@@ -594,8 +689,8 @@ impl QrState<Scanned> {
 }
 
 impl QrState<Cancelled> {
-    fn new(cancel_code: CancelCode) -> Self {
-        QrState { state: Cancelled::new(cancel_code) }
+    fn new(cancelled_by_us: bool, cancel_code: CancelCode) -> Self {
+        QrState { state: Cancelled::new(cancelled_by_us, cancel_code) }
     }
 
     fn as_content(&self, flow_id: &FlowId) -> OutgoingContent {
@@ -614,10 +709,10 @@ impl QrState<Created> {
                 if self.state.secret == m.secret {
                     Ok(QrState { state: Scanned {} })
                 } else {
-                    Err(QrState::<Cancelled>::new(CancelCode::KeyMismatch))
+                    Err(QrState::<Cancelled>::new(false, CancelCode::KeyMismatch))
                 }
             }
-            _ => Err(QrState::<Cancelled>::new(CancelCode::UnknownMethod)),
+            _ => Err(QrState::<Cancelled>::new(false, CancelCode::UnknownMethod)),
         }
     }
 }
@@ -627,7 +722,7 @@ impl QrState<Done> {
         self.state.as_content(flow_id)
     }
 
-    fn verified_identities(&self) -> (Arc<[ReadOnlyDevice]>, Arc<[UserIdentities]>) {
+    fn verified_identities(&self) -> (Arc<[ReadOnlyDevice]>, Arc<[ReadOnlyUserIdentities]>) {
         (self.state.verified_devices.clone(), self.state.verified_master_keys.clone())
     }
 }
@@ -637,7 +732,7 @@ impl QrState<Confirmed> {
         self,
         _: &DoneContent,
         verified_device: Option<&ReadOnlyDevice>,
-        verified_identity: Option<&UserIdentities>,
+        verified_identity: Option<&ReadOnlyUserIdentities>,
     ) -> QrState<Done> {
         let devices: Vec<_> = verified_device.into_iter().cloned().collect();
         let identities: Vec<_> = verified_identity.into_iter().cloned().collect();
@@ -676,7 +771,7 @@ impl QrState<Reciprocated> {
         self,
         _: &DoneContent,
         verified_device: Option<&ReadOnlyDevice>,
-        verified_identity: Option<&UserIdentities>,
+        verified_identity: Option<&ReadOnlyUserIdentities>,
     ) -> QrState<Done> {
         let devices: Vec<_> = verified_device.into_iter().cloned().collect();
         let identities: Vec<_> = verified_identity.into_iter().cloned().collect();
@@ -696,14 +791,14 @@ mod test {
 
     use matrix_qrcode::QrVerificationData;
     use matrix_sdk_test::async_test;
-    use ruma::identifiers::{event_id, room_id, user_id, DeviceIdBox, UserId};
+    use ruma::{event_id, room_id, user_id, DeviceIdBox, UserId};
 
     use crate::{
         olm::{PrivateCrossSigningIdentity, ReadOnlyAccount},
         store::{Changes, CryptoStore, MemoryStore},
         verification::{
             event_enums::{DoneContent, OutgoingContent, StartContent},
-            FlowId, IdentitiesBeingVerified,
+            FlowId, IdentitiesBeingVerified, VerificationStore,
         },
         QrVerification, ReadOnlyDevice,
     };
@@ -723,8 +818,10 @@ mod test {
     #[async_test]
     async fn test_verification_creation() {
         let store = memory_store();
-
         let account = ReadOnlyAccount::new(&user_id(), &device_id());
+
+        let store = VerificationStore { account: account.clone(), inner: store };
+
         let private_identity = PrivateCrossSigningIdentity::new(user_id()).await;
         let flow_id = FlowId::ToDevice("test_transaction".to_owned());
 
@@ -742,11 +839,12 @@ mod test {
         };
 
         let verification = QrVerification::new_self_no_master(
-            account,
             store.clone(),
             flow_id.clone(),
             master_key.clone(),
             identities.clone(),
+            false,
+            None,
         );
 
         assert_eq!(verification.inner.first_key(), &device_key);
@@ -758,6 +856,8 @@ mod test {
             master_key.clone(),
             device_key.clone(),
             identities.clone(),
+            false,
+            None,
         );
 
         assert_eq!(verification.inner.first_key(), &master_key);
@@ -775,6 +875,8 @@ mod test {
             master_key.clone(),
             bob_master_key.clone(),
             identities,
+            false,
+            None,
         );
 
         assert_eq!(verification.inner.first_key(), &master_key);
@@ -787,6 +889,8 @@ mod test {
             let alice_account = ReadOnlyAccount::new(&user_id(), &device_id());
             let store = memory_store();
 
+            let store = VerificationStore { account: alice_account.clone(), inner: store };
+
             let bob_account = ReadOnlyAccount::new(alice_account.user_id(), "BOBDEVICE".into());
 
             let private_identity = PrivateCrossSigningIdentity::new(user_id()).await;
@@ -813,15 +917,18 @@ mod test {
             };
 
             let alice_verification = QrVerification::new_self_no_master(
-                alice_account.clone(),
                 store,
                 flow_id.clone(),
                 master_key.clone(),
                 identities,
+                false,
+                None,
             );
 
             let bob_store = memory_store();
 
+            let bob_store = VerificationStore { account: bob_account.clone(), inner: bob_store };
+
             let mut changes = Changes::default();
             changes.identities.new.push(identity.into());
             changes.devices.new.push(alice_device.clone());
@@ -832,30 +939,32 @@ mod test {
 
             let bob_verification = QrVerification::from_scan(
                 bob_store,
-                bob_account,
                 private_identity,
                 alice_account.user_id().to_owned(),
                 alice_account.device_id().to_owned(),
                 flow_id,
                 qr_code,
+                false,
+                None,
             )
             .await
             .unwrap();
 
-            let content = bob_verification.reciprocate().unwrap();
+            let request = bob_verification.reciprocate().unwrap();
+            let content = OutgoingContent::try_from(request).unwrap();
             let content = StartContent::try_from(&content).unwrap();
 
             alice_verification.receive_reciprocation(&content);
 
             let request = alice_verification.confirm_scanning().unwrap();
-            let content = OutgoingContent::from(request);
+            let content = OutgoingContent::try_from(request).unwrap();
             let content = DoneContent::try_from(&content).unwrap();
 
             assert!(!alice_verification.is_done());
             assert!(!bob_verification.is_done());
 
             let (request, _) = bob_verification.receive_done(&content).await.unwrap();
-            let content = OutgoingContent::from(request.unwrap());
+            let content = OutgoingContent::try_from(request.unwrap()).unwrap();
             let content = DoneContent::try_from(&content).unwrap();
             alice_verification.receive_done(&content).await.unwrap();
 
@@ -871,8 +980,8 @@ mod test {
 
             let identity = identity.own().unwrap();
 
-            assert!(!bob_device.is_trusted());
-            assert!(alice_device.is_trusted());
+            assert!(!bob_device.is_locally_trusted());
+            assert!(alice_device.is_locally_trusted());
             assert!(identity.is_verified());
         };
 
